/*

  Pager Service Central Server for A0B36APO Subject
  Czech technical University in Prague
  Copyright 2014 - 2015, Pavel Pisa <pisa@cmp.felk.cvut.cz>

  Code based on uLAN ETHERNET daemon (ul_ethd)
  by Petr Smolik and Pavel Pisa and other GPL/MPL based
  projects.
*/

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <sys/types.h>
#ifdef _WIN32
 #include <winsock.h>
 #include <fcntl.h>
 typedef int socklen_t;
#else
 #include <sys/select.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #define closesocket(s)  close(s)
#endif /* _WIN32 */
#include <sys/stat.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>

#define HAS_GETOPT_LONG
#define HAS_GETDELIM

#include "utils.h"
#include "pagercentral.h"

/*******************************************************************/
/* Global variables */

const char *pagec_config_filename = NULL;
int pagec_server_port = 55556;
int pagec_server_socket;
int pagec_server_terminated;
int verbose_fl;

pagec_config_t *pagec_config;

pthread_mutex_t pagec_mutex = PTHREAD_MUTEX_INITIALIZER;

/*******************************************************************/
/* Utility and portability functions */

#ifndef HAS_GETDELIM
int getdelim(char **line,size_t *linelen,char delim,FILE *F)
{
 char c;
 size_t  l=0;
 do{
  if(l+1>=*linelen)
  {
   *linelen=l+20;
   if(!*line) *line=(char *)malloc(*linelen);
    else *line=(char *)realloc(*line,*linelen);
  }
  c=fgetc(F);
  if(feof(F)) {if(l) break; else return -1;}
  if(c!='\r') (*line)[l++]=c;
 } while(c!=delim);
 (*line)[l]=0;
 return l;
}
#endif

void line_rtrim(char *line)
{
  size_t l = strlen(line);
  while (l) {
    if(!strchr("\n\r \t", line[l - 1]))
      break;
    l--;
    line[l] = 0;
  }
}

/*******************************************************************/
/* Configuration file read */

int pagec_message_id_gen(pagec_config_t *cfg)
{
  if (cfg->mesage_id_last >= 1000000)
    cfg->mesage_id_last = 0;
  return ++cfg->mesage_id_last;
}

pagec_pager_t *pagec_config_pager_find(pagec_config_t *cfg, int id, int allow_alloc)
{
  pagec_pager_t *pager = cfg->pagers;

  while (pager) {
    if (pager->pager_id == id)
      return pager;
    pager=pager->next;
  }
  if (allow_alloc) {
    pager = malloc(sizeof(*pager));
    if (pager == NULL)
      return NULL;
    memset(pager, 0, sizeof(*pager));
    pager->pager_id = id;
    pager->messages_ptail = &pager->messages_head;
    pager->next = cfg->pagers;
    cfg->pagers = pager;
    return pager;
  }
  return NULL;
}

pagec_message_t *pagec_pager_message_find(pagec_pager_t *pager, int id, int unlink)
{
  pagec_message_t **pmessage = &pager->messages_head;
  pagec_message_t *message;

  while ((message = *pmessage) != NULL) {
    if (message->message_id == id) {
      if (unlink) {
        *pmessage = message->next;
        if (pager->messages_ptail == &message->next)
          pager->messages_ptail = pmessage;
      }
      return message;
    }
    pmessage = &message->next;
  }
  return NULL;
}

void pagec_pager_message_insert(pagec_pager_t *pager, pagec_message_t *message)
{
  *pager->messages_ptail = message;
  pager->messages_ptail = &message->next;
}


void pagec_config_destroy(pagec_config_t *cfg)
{
  pagec_pager_t *pager;
  pagec_message_t *message;
  void *p;

  if (cfg == NULL)
    return;

  pager = cfg->pagers;

  while(pager) {
    message = pager->messages_head;
    while (message) {
      if (message->message_text)
        free(message->message_text);
      p = message;
      message = message->next;
      free (p);
    }
    p = pager;
    pager = pager->next;
    free(p);
  }

  free(cfg);
}

#define PAGEC_CONFIG_TOKEN_MAX 20

pagec_config_t *pagec_config_read(pagec_config_t *cfg, const char *config_filename)
{
  char *p, *r;
  char *line = NULL;
  size_t line_len = 0;
  size_t text_len;
  int line_num = 0;
  FILE *file;
  char token[PAGEC_CONFIG_TOKEN_MAX+1];
  pagec_pager_t *pager;
  pagec_message_t *message = NULL;

  file = fopen(config_filename, "r");
  if (!file) {
    fprintf(stderr, "pagec_config_read: file \"%s\" open error - %s\n",
            config_filename, strerror(errno));
    goto error_file;
  }

  while(1) {
      line_num++;
      if(getdelim(&line, &line_len, '\n', file) == -1)
        break;
      if (line == NULL)
        break;
      line_rtrim(line);
      p=line;
      si_skspace(&p);
      if (!*p || (*p == '#'))
        continue;
      if (si_alnumn(&p, token, PAGEC_CONFIG_TOKEN_MAX) <= 0) {
        fprintf(stderr, "pagec_config_read: line %d does not start with token \"%s\"\n",
            line_num, line);
        goto error_parse;
      }
      if (!strcmp(token, "message")) {
        long id;
        si_skspace(&p);
        if (si_long(&p, &id, 0) <= 0) {
          fprintf(stderr, "pagec_config_read: pager ID in message command parse error at line %d\n",
                       line_num);
          goto error_parse;
        }
        pager = pagec_config_pager_find(cfg, id, 1);
        if (pager == NULL) {
          fprintf(stderr, "pagec_config_read: pager id %ld not available for message at line %d\n",
                           id, line_num);
          goto error_parse;

        }
        message = (pagec_message_t *)malloc(sizeof(*message));
        if (message == NULL) {
          fprintf(stderr, "pagec_config_read: out of memory\n");
          goto error_parse;
        }
        memset(message, 0, sizeof(*message));
        message->message_id = pagec_message_id_gen(cfg);

        si_skspace(&p);

        if (si_alnumn(&p, token, PAGEC_CONFIG_TOKEN_MAX) <= 0) {
          fprintf(stderr, "pagec_config_read: gate line %d garbage instead of message config\n",
                  line_num);
          goto error_parse;
        }
        if (strcmp(token, "text")) {
          fprintf(stderr, "pagec_config_read: message line %d \"text\" token missing\n",
                  line_num);
          goto error_parse;
        }
        si_skspace(&p);
        if (*p != '"') {
          fprintf(stderr, "pagec_config_read: message line %d text opening '\"' character missing\n",
                 line_num);
          goto error_parse;
        }
        p++;
        r = strchr(p, '"');
        if (r == NULL)  {
          fprintf(stderr, "pagec_config_read: message line %d text closing '\"' character missing\n",
                 line_num);
          goto error_parse;
        }
        text_len = r - p;
        message->message_text = malloc(text_len + 1);
        if (message->message_text == NULL) {
          fprintf(stderr, "pagec_config_read: out of memory\n");
          goto error_parse;
        }

        memcpy(message->message_text, p, text_len);
        message->message_text[text_len] = 0;

        pagec_pager_message_insert(pager, message);

        p = r + 1;
        si_skspace(&p);

        message = NULL;
      } else if (!strcmp(token, "pager")) {
        long id, pin;
        si_skspace(&p);
        if (si_long(&p, &id, 0) <= 0) {
          fprintf(stderr, "pagec_config_read: pager ID parse error at line %d\n",
                       line_num);
          goto error_parse;
        }
        pager = pagec_config_pager_find(cfg, id, 1);
        if (pager == NULL) {
          fprintf(stderr, "pagec_config_read: out of memory\n");
          goto error_parse;
        }

        si_skspace(&p);
        while (*p) {
          if (si_alnumn(&p, token, PAGEC_CONFIG_TOKEN_MAX) <= 0) {
            fprintf(stderr, "pagec_config_read: pager line %d garbage instead of pager config\n",
                    line_num);
            goto error_parse;
          }
          if (strcmp(token, "pin")) {
            fprintf(stderr, "pagec_config_read: pager %d line %d \"pin\" token missing\n",
                    pager->pager_id, line_num);
            goto error_parse;
          }
          si_skspace(&p);
          if (si_long(&p, &pin, 0) <= 0) {
            fprintf(stderr, "pagec_config_read: pager %d pin parse error at line %d\n",
                            pager->pager_id, line_num);
            goto error_parse;
          }
          pager->pin = pin;
          si_skspace(&p);
        }

        pager = NULL;
      } else {
        fprintf(stderr, "pagec_config_read: unknown token \"%s\" at line %d\n",
            token, line_num);
        goto error_parse;
      }
  }

  fclose(file);

  if (line != NULL)
    free(line);

  return cfg;

error_parse:
  if (message != NULL) {
    if (message->message_text != NULL)
      free(message->message_text);
    free(message);
  }

error_file:
  pagec_config_destroy(cfg);

  if (line != NULL)
    free(line);

  return NULL;
}

/*******************************************************************/
/* Clients requests processing */

#define PAGEC_CLIENT_TOKEN_MAX 20

void* pagec_client_run(void *vcontext)
{
  pagec_client_context_t *client=(pagec_client_context_t *)vcontext;
  pagec_config_t *cfg = pagec_config;
  int fd = client->sock;
  char token[PAGEC_CLIENT_TOKEN_MAX+1];
  pagec_pager_t *pager = NULL;
  pagec_message_t *message = NULL;
  char *p;
  char *line = NULL;
  size_t line_len = 0;
  int r = -1;
  int res;

  if (verbose_fl)
    fprintf(stderr, "pagercentral: thread for client fd:%d is running\n",fd);

  pthread_mutex_lock (&pagec_mutex);

  while (1) {

    if (client->terminated || feof(client->sock_file)) {
      r=0;
      break;
    }

    /* simple - coarse data structures locking */
    pthread_mutex_unlock (&pagec_mutex);

    fflush(client->sock_file);
    res = getdelim(&line, &line_len, '\n', client->sock_file);

    pthread_mutex_lock (&pagec_mutex);

    if(res == -1) {
      if (feof(client->sock_file))
        continue;
      fprintf(stderr, "pagercentral: client fd:%d getdelim error %s\n",
                      fd, strerror(errno));
      break;
    }
    if (line == NULL) {
      fprintf(stderr, "pagercentral: client fd:%d getdelim empty line\n", fd);
      break;
    }
    line_rtrim(line);
    p=line;
    si_skspace(&p);
    if (!*p || (*p == '#'))
      continue;

    if (verbose_fl)
      fprintf(stderr, "pagercentral: client fd:%d request \"%s\"\n", fd, line);

    if (si_alnumn(&p, token, PAGEC_CLIENT_TOKEN_MAX) <= 0) {
      fprintf(stderr, "pagercentral: client fd:%d request missing token \"%s\"\n",
                      fd, line);
      fprintf(client->sock_file, "error hint \"request missing token\"\n");
      break;
    }

    if (!strcmp(token, "querrymessage")) {
      long pager_id;

      if (si_long(&p, &pager_id, 0) <= 0) {
        fprintf(stderr, "pagercentral: client fd:%d querrymessage request missing pager ID\n",
                         fd);
        fprintf(client->sock_file, "error hint \"missing pager ID\"\n");
        continue;
      }

      pager = pagec_config_pager_find(cfg, pager_id, 1);
      if (pager == NULL) {
        fprintf(stderr, "pagercentral: client fd:%d querrymessage pager ID %ld unknown\n",
                       fd, pager_id);
        fprintf(client->sock_file, "querrymessage error hint \"pager id %ld unknown\"\n",
                       pager_id);
        continue;
      }

      message = pager->messages_head;
      if (message != NULL) {
        fprintf(client->sock_file, "querrymessage_r %d %d %d %s\n", pager->pager_id,
                message->message_id, message->sender_id,
                message->message_text? message->message_text: "");
      } else {
        fprintf(client->sock_file, "querrymessage_r %d none\n", pager->pager_id);
      }
    } else if (!strcmp(token, "confirmmessage")) {
      long message_id;
      long pager_id;

      if (si_long(&p, &pager_id, 0) <= 0) {
        fprintf(stderr, "pagercentral: client fd:%d confirmmessage request missing pager ID\n",
                         fd);
        fprintf(client->sock_file, "error hint \"missing pager ID\"\n");
        continue;
      }

      if (si_long(&p, &message_id, 0) <= 0) {
        fprintf(stderr, "pagercentral: client fd:%d confirmmessage request missing message ID\n",
                         fd);
        fprintf(client->sock_file, "error hint \"missing message ID\"\n");
        continue;
      }

      pager = pagec_config_pager_find(cfg, pager_id, 0);
      if (pager == NULL) {
        fprintf(stderr, "pagercentral: client fd:%d confirmmessage pager ID %ld unknown\n",
                       fd, pager_id);
        fprintf(client->sock_file, "confirmmessage error hint \"pager id %ld unknown\"\n",
                       pager_id);
        continue;
      }

      message = pagec_pager_message_find(pager, message_id, 1);
      if (message == NULL) {
        fprintf(stderr, "pagercentral: client fd:%d confirmmessage message ID %ld unknown\n",
                       fd, message_id);
        fprintf(client->sock_file, "confirmmessage error hint \"message id %ld unknown\"\n",
                       message_id);
        continue;
      }

      if (message->message_text)
        free(message->message_text);
      free(message);

      fprintf(client->sock_file, "confirmmessage_r %ld %ld\n", pager_id, message_id);
    } else if (!strcmp(token, "sendmessage")) {
      long pager_id;
      long recipient_id;
      size_t text_len;

      if (si_long(&p, &pager_id, 0) <= 0) {
        fprintf(stderr, "pagercentral: client fd:%d sendmessage request missing source pager ID\n",
                         fd);
        fprintf(client->sock_file, "error hint \"missing pager ID\"\n");
        continue;
      }

      if (si_long(&p, &recipient_id, 0) <= 0) {
        fprintf(stderr, "pagercentral: client fd:%d sendmessage request missing recipient ID\n",
                         fd);
        fprintf(client->sock_file, "error hint \"missing recipient ID\"\n");
        continue;
      }

      pager = pagec_config_pager_find(cfg, recipient_id, 1);
      if (pager == NULL) {
        fprintf(stderr, "pagercentral: client fd:%d sendmessage recipient ID %ld unknown\n",
                       fd, pager_id);
        fprintf(client->sock_file, "sendmessage error hint \"recipient id %ld unknown\"\n",
                       pager_id);
        continue;
      }

      message = (pagec_message_t *)malloc(sizeof(*message));
      if (message == NULL) {
        fprintf(stderr, "pagercentral: client fd:%d sendmessage out of memory\n", fd);
        fprintf(client->sock_file, "sendmessage error out of memory\n");
        continue;
      }
      memset(message, 0, sizeof(*message));
      message->message_id = pagec_message_id_gen(cfg);
      message->sender_id = pager_id;

      si_skspace(&p);
      text_len = strlen(p);

      message->message_text = (char *)malloc(text_len + 1);
      if (message->message_text == NULL) {
        fprintf(stderr, "pagercentral: client fd:%d sendmessage out of memory\n", fd);
        fprintf(client->sock_file, "sendmessage error out of memory\n");
        free(message);
        continue;
      }
      strcpy(message->message_text, p);

      pagec_pager_message_insert(pager, message);

      fprintf(client->sock_file, "sendmessage_r %ld %ld %d\n", pager_id,
              recipient_id, message->message_id);
    } else {
      fprintf(stderr, "pagercentral: client fd:%d send unknown token \"%s\"\n",
                      fd, token);
      fprintf(client->sock_file, "error hint \"unknown token %s\"\n", token);
      continue;
    }
  }
  pthread_mutex_unlock (&pagec_mutex);

  if (verbose_fl)
    fprintf(stderr, "pagercentral: thread for client fd:%d terminated by rcode:%d\n",fd,r);

  fflush(client->sock_file);
  fclose(client->sock_file);

  if (r==-1) {
    if (verbose_fl)
      fprintf(stderr, "pagercentral: closing socket in thread\n");

    /* closesocket(client->sock); */
    /* cannot be called, socket alread closed by fclose, but windows are question */

    /* terminated by client side request */
    /*
    pthread_mutex_lock (&client_mutex);
    client_delete(&client_root,client);
    pthread_mutex_unlock (&client_mutex);
    */
  }
  free(client);

  if (line)
    free(line);

  if (verbose_fl)
    fprintf(stderr, "pagercentral: thread for client fd:%d was terminated\n",fd);

  pthread_detach(pthread_self());
  pthread_exit(NULL);
  return NULL;
}


/*******************************************************************/
/* Remote requests server */

int pagec_server_init(void)
{
  struct sockaddr_in address;
  int server_socket;
  int yes=1;
 #ifdef _WIN32
  WORD wVersionRequested;
  WSADATA wsaData;
  int sockopt = SO_SYNCHRONOUS_NONALERT;
 #endif

  pagec_server_terminated = 0;

 #ifdef _WIN32
  wVersionRequested = MAKEWORD(2, 0);
  WSAStartup(wVersionRequested, &wsaData);
  setsockopt(INVALID_SOCKET, SOL_SOCKET, SO_OPENTYPE, (char *)&sockopt, sizeof(sockopt));
 #endif

  if ((server_socket = socket(AF_INET,SOCK_STREAM,0)) < 0) {
    fprintf(stderr, "pagecental: the socket created error\n");
    exit(1);
  }
  if (verbose_fl)
    fprintf(stderr, "pagecental: the server socket fd:%d was created\n", server_socket);

  /* Make listening socket's port reusable - must appear before bind */
  if (setsockopt(server_socket, SOL_SOCKET, SO_REUSEADDR, (char*)&yes,sizeof(yes)) < 0) {
    fprintf(stderr, "pagecental: setsockopt failure\n");
    exit(1);
  }

  address.sin_family = AF_INET;
  address.sin_addr.s_addr = INADDR_ANY;
  address.sin_port = htons(pagec_server_port);
  if (bind(server_socket,(struct sockaddr *)&address,sizeof(address)) < 0) {
    fprintf(stderr, "pagecental: bind error\n");
    exit(1);
  }
  if (verbose_fl)
    fprintf(stderr, "pagecental: server socket fd:%d bound to port %d \n", server_socket, pagec_server_port);

  /*  Create a connection queue and initialize readfds to handle input from server_socket. */
  if(listen(server_socket, 5) < 0) {
    fprintf(stderr, "pagecental: error to listen on socket\n");
    exit(1);
  }

  pagec_server_socket = server_socket;

  return 0;
}

int pagec_server_run(void)
{
  int csock;
  fd_set readfds, testfds;
  int result, f;
  socklen_t addrlen;
  struct sockaddr_in address;
  int server_socket = pagec_server_socket;

  if (verbose_fl)
    fprintf(stderr, "pagecental: server run\n");

  FD_ZERO(&readfds);
  FD_SET(server_socket, &readfds);

  while(!pagec_server_terminated) {
    struct timeval timeout = {.tv_sec = 5, .tv_usec = 0};

    testfds=readfds;
    result = select(FD_SETSIZE, &testfds, (fd_set *)0, (fd_set *)0, &timeout);

    if(result < 0) {
      break;
    }

    if (result == 0)
      continue;

    if(FD_ISSET(server_socket, &testfds)) { /*  request for a new connection */
      addrlen = sizeof(struct sockaddr_in);
      csock = accept(server_socket, (struct sockaddr *)&address, &addrlen);
      if (csock > 0){
        pagec_client_context_t *client;
        if (verbose_fl)
          fprintf(stderr, "pagecental: the client from %s is connected, fd:%d\n",
                          inet_ntoa(address.sin_addr), csock);

        client = (pagec_client_context_t *)malloc(sizeof(pagec_client_context_t));
        if (client == NULL) {
          fprintf(stderr, "pagecental: could not get memory\n");
          exit(1);
        }
        memset(client, 0, sizeof(*client));
        client->sock = csock;
        client->terminated = 0;

       #ifndef _WIN32
        f = client->sock;
       #else
        f = _open_osfhandle(client->sock, _O_RDWR);
        if (f < 0) {
          fprintf(stderr, "pagecental: error _open_osfhandle\n");
          exit(1);
        }
       #endif /* _WIN32 */

        client->sock_file = fdopen(f, "r+");
        if (!client->sock_file) {
          fprintf(stderr, "pagecental: error fdopen fd:%d\n", client->sock);
          exit(1);
        }

        /*
        pthread_mutex_lock (&client_mutex);
        client_insert(&client_root, client);
        pthread_mutex_unlock (&client_mutex);
        */

        if(pthread_create(&client->threadid, NULL, pagec_client_run, (void*)client) != 0){
          fprintf(stderr, "pagecental: could not create client thread\n");
          exit(1);
        }
      }
    } else {
      // this should never happen if the program work well
      fprintf(stderr, "pagecental: ERROR: read from unregistered fd\n");
      exit(1);
    }
  }

  return 0;
}

void pagec_server_stop(void)
{
  pagec_server_terminated = 1;
}

/*******************************************************************/
/* Program main/startup/clenup and command line parsing */

void
sigexit(int sig)
{
  fprintf(stderr, "pagecental: signal %d received\n", sig);
  pagec_server_stop();
}

static void
usage(void)
{
  printf("Usage: pagercentral <parameters>\n");
  printf("  -c, --config <filename>  n device [none]\n");
  printf("  -p, --port <value>       server port [%d]\n", pagec_server_port);
  printf("  -v, --verbose            verbose output\n");
  printf("  -V, --version            show version\n");
  printf("  -h, --help               this usage screen\n");
}

int main(int argc,char *argv[])
{
  pagec_config_t *cfg;

  static struct option long_opts[] = {
    { "config", 1, 0, 'c' },
    { "port", 1, 0, 'p' },
    { "verbose", 0, 0, 'v' },
    { "version",0,0, 'V' },
    { "help",  0, 0, 'h' },
    { 0, 0, 0, 0}
  };
  int opt;

 #ifndef HAS_GETOPT_LONG
  while ((opt = getopt(argc, argv, "c:p:vVh")) != EOF)
 #else
  while ((opt = getopt_long(argc, argv, "c:p:vVh",
                            &long_opts[0], NULL)) != EOF)
 #endif
    switch (opt) {
    case 'c':
      pagec_config_filename = optarg;
      break;
    case 'p':
      pagec_server_port = strtol(optarg,NULL,0);
      break;
    case 'v':
      verbose_fl = 1;
      break;
    case 'V':
      fputs("pagercentral 0.1 (APO 2015)\n", stdout);
      exit(0);
    case 'h':
    default:
      usage();
      exit(opt == 'h' ? 0 : 1);
    }


  cfg = (pagec_config_t *)malloc(sizeof(*cfg));
  if (cfg == NULL) {
    fprintf(stderr, "pagec_config_read: out of memory\n");
    exit(1);;
  }
  memset(cfg, 0, sizeof(*cfg));
  pagec_config = cfg;

  if (pagec_config_filename != NULL) {
    pagec_config = pagec_config_read(cfg, pagec_config_filename);
    if (pagec_config == NULL) {
      fprintf(stderr, "pagercentral: gonfig file \"%s\" not parsed\n",
            pagec_config_filename);
      exit(1);
    }
  }

  pagec_server_init();
  signal(SIGINT,  sigexit);
  signal(SIGTERM,  sigexit);
  pagec_server_run();
  if (verbose_fl)
    fprintf(stderr, "pagecental: program finishing\n");
  pagec_config_destroy(pagec_config);
  return 0;
}
